local L = AceLibrary("AceLocale-2.2"):new("GhostPulse")

local CHECK_INTERVAL = 0.25

AuraWatcher = {

	New = function(this, env, listenerFunction)
		-- check arguments
		if not env or type(env) ~= "table" then
			error("Must supply an environment tracker to AuraWatcher:New()")
		end
		
		if not listenerFunction or type(listenerFunction) ~= "function" then
			error("Must supply a listener function to AuraWatcher:Connect()")
		end
			
		-- globals
		frame = CreateFrame("Frame")
		
		frame.nextCheckTime = 0
		
		frame.enabled = false		
		frame.enabledAlerts = { }

		frame.buffCache = { }
		frame.debuffCache = { }
		frame.textureCache = { }
		
		frame.environment = env
		frame.OnBuffGain = listenerFunction
		frame.OnBuffFade = listenerFunction
		frame.OnDebuffGain = listenerFunction
		frame.OnDebuffFade = listenerFunction
		
		-- functions
		frame:RegisterEvent("UNIT_AURA")
		
		frame.GetAlertTypes = function()
			return {
				["buffgain"] = L["Player Buff Gains"],
				["bufffade"] = L["Player Buff Fades"],
				["debuffgain"] = L["Player Debuff Gains"],
				["debufffade"] = L["Player Debuff Fades"],
			}
		end
		
		-- common configuration
		frame.SetOptions = function(this, options)
			this.settings = options
		end
		
		frame.CheckEnv = function(this, alertType)
			return this:CheckCombat(alertType) and this:CheckGrouping(alertType) and
				   ( this.settings[alertType].whileMounted == true or not IsMounted() )
		end
		
		frame.CheckCombat = function(this, alertType)
			local rc = true
			
			
			if this.settings and this.settings[alertType].inCombat then
				rc = this.environment:PlayerIsInCombat()
			end
			
			return rc
		end

		frame.CheckGrouping = function(this, alertType)
			local rc = true
			local curGroupType = this.environment:GetPlayerGroupType()
			
			
			if curGroupType == "solo" then
				rc = this.settings[alertType].solo
			elseif curGroupType == "party" then
				rc = this.settings[alertType].party
			elseif curGroupType == "raid" then
				rc = this.settings[alertType].raid
			end
			
			return rc
		end
		
		frame.CanShowThing = function(this, alertType, thing)
			local rc = true
			
			
			if thing and alertType and this.settings[alertType] then
				if not this.enabledAlerts[alertType] then
					rc = false
				else
					if this.settings[alertType].whitelist and table.isEmpty(this.settings[alertType].whitelist) == false then
						rc = this.settings[alertType].whitelist[thing]
					else
						if this.settings[alertType].blacklist then
							if this.settings[alertType].blacklist[thing] then
								rc = false
							end
						end
					end
				end
			end
		
			return rc
		end
		
		frame.SetAlertEnabled = function(this, alertType, value)
			this.enabledAlerts[alertType] = value
		end
		
		-- watcher-specific function
		frame:SetScript("OnEvent", function(this, event, unit)
			if this.enabled and event == "UNIT_AURA" and unit == "player" then
				this.nextCheckTime = GetTime() + CHECK_INTERVAL
				this:SetScript("OnUpdate", this.Scan)
			end
		end)
		
		frame.Scan = function(this)
			local curTime = GetTime()
			
			
			if curTime >= this.nextCheckTime then
				if this.enabledAlerts["buffgain"] or this.enabledAlerts["bufffade"] then
					this:ScanBuffs()
				end
				
				if this.enabledAlerts["debuffgain"] or this.enabledAlerts["debufffade"] then
					this:ScanDebuffs()
				end
				
				this:SetScript("OnUpdate", nil)
			end
		end
		
		frame.ScanBuffs = function(this)
			local recentBuffs = { }


			-- check for new buffs
			for i = 1, 40 do
				local id = GetPlayerBuff(i, "HELPFUL")
				
				
				if id > 0 then
					local name = GetPlayerBuffName(id)
					local texture = GetPlayerBuffTexture(id)
					
					
					if name and texture then
						recentBuffs[name] = true
						this.textureCache[name] = this.textureCache[name] or texture
						
						if this:CanShowThing("buffgain", name) then
							if this:CheckEnv("buffgain") then
								-- was this buff here last time we checked?
								if not this.buffCache[name]  then
									-- alert
									if this.OnBuffGain then
										if type(this.OnBuffGain) == "function" then
											this:OnBuffGain("buffgain", name, this.textureCache[name], {
												["Unit"] = "player",
												["Name"] = name,
												["Type"] = "buff",
												["Subtype"] = "gain",
											})
										else
											error("AuraWatcher.OnBuffGain must be a function")
										end
									end
								end
							end
						end
					end
				end
			end
			
			-- check for buffs that have faded
			if this:CheckEnv("bufffade") then
				for i, _ in pairs(this.buffCache) do
					if this:CanShowThing("bufffade", i) then
						if recentBuffs[i] == nil then
							-- alert
							if this.OnBuffFade then
								if type(this.OnBuffFade) == "function" then
									this:OnBuffFade("bufffade", i, this.textureCache[i], {
										["Unit"] = "player",
										["Name"] = i,
										["Type"] = "buff",
										["Subtype"] = "fade",
									})
								else
									error("AuraWatcher.OnBuffFade must be a function")
								end
							end
						end
					end
				end
			end
						
			this.buffCache = recentBuffs
		end
		
		frame.ScanDebuffs = function(this)
			local recentDebuffs = { }

			for i = 1, 40 do
				local name, _, texture, count = UnitDebuff("player", i)
				
				if name and texture then
					count = count or 0
					recentDebuffs[name] = count
					this.textureCache[name] = this.textureCache[name] or texture
						
					if this:CanShowThing("debuffgain", name) then
						if this:CheckEnv("debuffgain") then
							if this.debuffCache[name] == nil or this.debuffCache[name] < count then
								-- alert
								if this.OnDebuffGain then
									if type(this.OnDebuffGain) == "function" then
										this:OnDebuffGain("debuffgain", name, this.textureCache[name], {
											["Unit"] = "player",
											["Name"] = name,
											["Type"] = "debuff",
											["Subtype"] = "gain",
											["Count"] = count,
										})
									else
										error("AuraWatcher.OnDebuffGain must be a function")
									end
								end
							end
						end
					end
				else
					break
				end
			end
			
			-- check for debuffs falling off
			if this:CheckEnv("debufffade") then
				for i, _ in pairs(this.debuffCache) do
					if this:CanShowThing("debufffade", i) then
						if recentDebuffs[i] == nil then
							-- alert
							if this.OnDebuffFade then
								if type(this.OnDebuffFade) == "function" then
									this:OnDebuffFade("debufffade", i, this.textureCache[i], {
										["Unit"] = "player",
										["Name"] = i,
										["Type"] = "debuff",
										["Subtype"] = "fade",
									})
								else
									error("AuraWatcher.OnDebuffFade must be a function")
								end
							end
						end
					end
				end
			end
			
			this.debuffCache = recentDebuffs
		end
		
		return frame
	end,
	
}