local L = AceLibrary("AceLocale-2.2"):new("GhostPulse")

PlayerDeathWatcher = {

	New = function(this, env, listenerFunction)
		-- check arguments
		if not env or type(env) ~= "table" then
			error("Must supply an environment tracker to PlayerDeathWatcher:New()")
		end
		
		if not listenerFunction or type(listenerFunction) ~= "function" then
			error("Must supply a listener function to PlayerDeathWatcher:New()")
		end

		-- globals
		frame = CreateFrame("Frame")
		
		frame.enabled = false
		frame.enabledAlerts = { }
				
		frame.lastKnownAlive = { }
		
		frame.environment = env
		frame.OnPlayerDeath = listenerFunction
		frame.OnPlayerResurrection = listenerFunction

		frame.GetAlertTypes = function()
			return {
				["playerdeath"] = L["Player Deaths"],
				["playerresurrection"] = L["Player Resurrections"],
			}
		end
		
		-- functions
		frame:RegisterEvent("UNIT_HEALTH")
		frame:RegisterEvent("RAID_ROSTER_UPDATE")
		frame:RegisterEvent("PARTY_MEMBERS_CHANGED")
		
		frame:SetScript("OnEvent", function(this, event, ...)
			if this.enabled then
				if event == "UNIT_HEALTH" then
					this:CheckUnitHealth(...)
				elseif event == "RAID_ROSTER_UPDATE" or event == "PARTY_MEMBERS_CHANGED" then
					this:ScanGroupHealth()
				end
			end
		end)
		
		-- common configuration
		frame.SetOptions = function(this, options)
			this.settings = options
		end
		
		frame.CheckEnv = function(this, alertType)
			return this:CheckCombat(alertType) and this:CheckGrouping(alertType) and
				   ( this.settings[alertType].whileMounted == true or not IsMounted() )
		end
		
		frame.CheckCombat = function(this, alertType)
			local rc = true
			
			
			if this.settings and this.settings[alertType].inCombat then
				rc = this.environment:PlayerIsInCombat()
			end
			
			return rc
		end

		frame.CheckGrouping = function(this, alertType)
			local rc = true
			local curGroupType = this.environment:GetPlayerGroupType()
			
			
			if curGroupType == "solo" then
				rc = this.settings[alertType].solo
			elseif curGroupType == "party" then
				rc = this.settings[alertType].party
			elseif curGroupType == "raid" then
				rc = this.settings[alertType].raid
			end
			
			return rc
		end
		
		frame.CanShowThing = function(this, alertType, thing)
			local rc = true
			
			
			if thing and alertType and this.settings[alertType] then
				if not this.enabledAlerts[alertType] then
					rc = false
				else
					if this.settings[alertType].whitelist and table.isEmpty(this.settings[alertType].whitelist) == false then
						rc = this.settings[alertType].whitelist[thing]
					else
						if this.settings[alertType].blacklist then
							if this.settings[alertType].blacklist[thing] then
								rc = false
							end
						end
					end
				end
			end
		
			return rc
		end
		
		frame.SetAlertEnabled = function(this, alertType, value)
			this.enabledAlerts[alertType] = value
		end
		
		-- watcher-specific function
		frame.CheckUnitHealth = function(this, unit)
			if this.InterestedInUnitHealth(unit) then
				local name = UnitName(unit)
				local _, class = UnitClass(unit)
				
				if class then
					local texture = "Interface\\AddOns\\Ghost_CDP\\images\\"..class..".tga"
					
			
					if not UnitIsDead(unit) and not UnitIsGhost(unit) then
						if this.enabledAlerts["playerresurrection"] and this:CheckEnv("playerresurrection") then
							-- alive
							if not this.lastKnownAlive[name] then
								-- alert
								if this.OnPlayerResurrection then
									if type(this.OnPlayerResurrection) == "function" then
										this:OnPlayerResurrection("playerresurrection", UnitName(unit), texture, {
											["Unit"] = unit,
											["Name"] = UnitName(unit),
											["Type"] = "playerresurrection",
										})
									else
										error("PlayerDeathWatcher.OnPlayerResurrection must be a function")
									end
								end
							end
						end
												
						this.lastKnownAlive[name] = true
					else
						if this.enabledAlerts["playerdeath"] and this:CheckEnv("playerdeath") then
							-- dead
							if this.lastKnownAlive[name] then
								-- alert
								if this.OnPlayerDeath then
									if type(this.OnPlayerDeath) == "function" then
										this:OnPlayerDeath("playerdeath", UnitName(unit), texture, {
											["Unit"] = unit,
											["Name"] = UnitName(unit),
											["Type"] = "playerdeath",
										})
									else
										error("PlayerDeathWatcher.OnPlayerDeath must be a function")
									end
								end
							end
						end
																	
						this.lastKnownAlive[name] = false
					end
				end
			end
		end
		
		frame.ScanGroupHealth = function(this)
			local prefix
			local upper
			
			
			this.lastKnownAlive = { }
			
			if GetNumRaidMembers() > 0 then
				prefix = "raid"
				upper = 40
			elseif GetNumPartyMembers() > 0 then
				prefix = "party"
				upper = 5
			else
				prefix = nil
				upper = nil
			end
			
			if prefix then
				for i = 1, upper do
					local unit = prefix..i
					
					
					if UnitExists(unit) then
						local name = UnitName(unit)
						
						
						if not UnitIsDead(unit) and not UnitIsGhost(unit) then
							this.lastKnownAlive[name] = true
						else
							this.lastKnownAlive[name] = false
						end
					end
				end
			end
			
			this.lastKnownAlive[UnitName("player")] = not UnitIsDead("player") and not UnitIsGhost("player")
		end

		frame.InterestedInUnitHealth = function(unit)
			local interest = false
			
			
			if unit then
				if unit == "player" then
					interest = true
				else
					if string.sub(unit, 1, 5) == "party" then
						interest = true
					elseif string.sub(unit, 1, 4) == "raid" then
						interest = true
					end
					
					if interest then
						if string.find(unit, "pet") then
							interest = false
						end
					end
				end
			end
			
			return interest
		end
			
		frame:ScanGroupHealth()
		
		return frame
	end,
	
}